/*
 * fltdiff.c
 * 
 * diff for files of binary floats returns proportional error.
 * (for fuzzy comparison of rasta versions)
 * 
 * 1997aug31 dpwe@icsi.berkeley.edu
 * $Header: /u/drspeech/src/rasta/testsuite/RCS/fltdiff.c,v 1.2 1998/03/18 02:58:32 dpwe Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>

#include "byteswap.h"

#define DFLT_TOL 0.001

int quiet = 0;
int silent = 0;
float maxerr = DFLT_TOL;
char *file1 = NULL;
char *file2 = NULL;
int cmppts = -1;
int hskip = 0;
int h1skip = 0;
int ascii = 0;
int debug = 0;

static void usage(char *name) {
    fprintf(stderr, "usage: %s [-q] [-s] [-A] [-m val] [-n n] [-k n] [-K n] reffile candfile\n", name);
    fprintf(stderr, "  compares the floating-point values in candfile with reffile\n");
    fprintf(stderr, "  and returns a nonzero exit code if their rms proportional\n");
    fprintf(stderr, "  difference is too large\n");
    fprintf(stderr, " -q    quiet processing (just set error code)\n");
    fprintf(stderr, " -s    silent processing (dummy for cmp(1) emulation)\n");
    fprintf(stderr, " -A    input files are CONFORMAL ascii, not binary floats\n");
    fprintf(stderr, " -m    acceptable rms error as proportion of file1 (%f)\n", DFLT_TOL);
    fprintf(stderr, " -n    total data points to compare (-1 => to EOF)\n");
    fprintf(stderr, " -k    header bytes to skip on both files (0)\n");
    fprintf(stderr, " -K    header bytes to skip on file1 (file2 if -ve) (0)\n");
    fprintf(stderr, " -d    debug output\n");
    exit(1);
}

#ifndef MAX
#define MAX(a,b)	((a)>(b))?(a):(b)
#define MIN(a,b)	((a)<(b))?(a):(b)
#endif /* MAX */

double FltBlkRMSDiff(float *buf1, float *buf2, int size)
{   /* calculate the RMS proportional difference between two blocks of floats.
       buf1 is the denominator. */
    double diff = 0, dd;
    long c = 0;
    int i;

    for (i=0; i < size; ++i) {
	if (*buf1 != 0) {
	    dd = (*buf1 - *buf2)/(*buf1);
	    diff += dd*dd;
	    ++c;
	}
    }
    if (c==0) c = 1;
    return sqrt(diff/c);
}

int read_ascii_line(float *buf, int maxcount, FILE *file)
{   /* read a line of text from file and parse it into no more than 
       <count> entries in the buf array.  Return how many it is. */
    char cbuf[8192];
    char *s, *t;
    int done = 0, count = 0;

    if(fgets(cbuf, 8192, file) != NULL) {
        s = cbuf;
	while (!done) {
	    buf[count] = strtod(s, &t);
	    if(t > s) {
		s = t;
		++count;
		if(count >= maxcount) {
		    fprintf(stderr, "read_ascii_line: overflowed float buffer of %d\n", maxcount);
		    done = 1;
		}
	    } else {
		done = 1;
	    }
	}
    }
    return count;
}

int main(int argc, char **argv) {
    char *programName = argv[0];
    FILE *f1;
    FILE *f2;
    float *b1;
    float *b2;
    int i, err = 0;
    int bufsize = 16384;
    int skip, got1, got2, thistime;
    int tot1 = 0, tot2 = 0;
    double rmse = 0;
    long count = 0;

    /* Handle the args */
    for (i = 1; i < argc; ++i) {
	if(argv[i][0] == '-') {
	    switch(argv[i][1]) {
	    case 'q':	quiet = 1; break;
	    case 's':	silent = 1; break;
	    case 'A':	ascii = 1; break;
	    case 'd':	debug = 1; break;
	    case 'm':	maxerr = atof(argv[++i]); break;
	    case 'n':	cmppts = atoi(argv[++i]); break;
	    case 'k':	hskip = atoi(argv[++i]); break;
	    case 'K':   h1skip = atoi(argv[++i]); break;
	    default:	++err;
	    }
	} else if (file1 == NULL) {
	    file1 = argv[i];
	} else if (file2 == NULL) {
	    file2 = argv[i];
	} else {
	    ++err;
	}
    }
    err += (file1 == NULL || file2 == NULL);
    if (err) {
	usage(programName);
    }
    if (debug) {
	fprintf(stderr, "%s: q=%d s=%d A=%d d=%d m=%f n=%d k=%d K=%d\n", 
		programName, quiet, silent, ascii, debug, maxerr, 
		cmppts, hskip, h1skip);
	fprintf(stderr, "ref='%s' cand='%s'\n", file1, file2);
    }

    if ( (f1 = fopen(file1, "r")) == NULL) {
	fprintf(stderr, "%s: couldn't read '%s'\n", file1);
	exit(-1);
    }
    if ( (f2 = fopen(file2, "r")) == NULL) {
	fprintf(stderr, "%s: couldn't read '%s'\n", file1);
	exit(-1);
    }

    /* skip to the appropriate places */
    skip = hskip + MAX(h1skip, 0);
    if (skip) {
	fseek(f1, skip, SEEK_SET);
    }
    skip = hskip + MAX(-h1skip, 0);
    if (skip) {
	fseek(f2, skip, SEEK_SET);
    }

    /* allocate buffers */
    assert(b1 = (float*)malloc(sizeof(float)*bufsize));
    assert(b2 = (float*)malloc(sizeof(float)*bufsize));

    thistime = bufsize;
    got1 = -1;
    while (got1 != 0 && cmppts != 0) {
	if (cmppts > 0) {
	    thistime = MIN(bufsize, cmppts);
	}
	if (ascii) {
	    got1 = read_ascii_line(b1, thistime, f1);
	    got2 = read_ascii_line(b2, thistime, f2);
	} else {
	    got1 = fread_swap(b1, sizeof(float), thistime, f1, 0);
	    got2 = fread_swap(b2, sizeof(float), thistime, f2, 0);
	}
	tot1 += got1;
	tot2 += got2;
	if (tot1 != tot2) {
	    fprintf(stderr, "file sizes differ at pt %d in %s and %d in %s\n",
		    tot1, file1, tot2, file2);
	    exit(-1);
	}
	rmse += got1 * pow(FltBlkRMSDiff(b1, b2, got1), 2);
	count += got1;
	if (cmppts > 0)	cmppts -= got1;
    }
    /* close the files */
    fclose(f1);
    fclose(f2);
    /* calculate the grand rms error */
    rmse = sqrt(rmse/count);
    /* report difference */
    if (!quiet) {
	printf("%s: ref=%s cand=%s\nrmspropdiff=%f\n", programName, 
	       file1, file2, rmse);
    }
    /* .. and set output code dep on "max" */
    if (rmse <= maxerr) {
	exit(0);
    } else {
	exit(1);
    }
}


